# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

import re

import pytest

from power_grid_model_io.utils.parsing import (
    is_node_ref,
    parse_node_ref,
    parse_trafo3_connection,
    parse_trafo_connection,
)


def test_parse_trafo_connection__pos():
    assert parse_trafo_connection("Dyn") == {"winding_from": "D", "winding_to": "yn", "clock": None}
    assert parse_trafo_connection("Yyn") == {"winding_from": "Y", "winding_to": "yn", "clock": None}
    assert parse_trafo_connection("Yzn") == {"winding_from": "Y", "winding_to": "zn", "clock": None}
    assert parse_trafo_connection("YNy") == {"winding_from": "YN", "winding_to": "y", "clock": None}
    assert parse_trafo_connection("Dy5") == {"winding_from": "D", "winding_to": "y", "clock": "5"}
    assert parse_trafo_connection("Dy11") == {"winding_from": "D", "winding_to": "y", "clock": "11"}
    assert parse_trafo_connection("Dyn-1") == {"winding_from": "D", "winding_to": "yn", "clock": "-1"}


def test_parse_trafo_connection__neg():
    with pytest.raises(ValueError, match="Invalid transformer connection string: 'Xyn'"):
        parse_trafo_connection("Xyn")
    with pytest.raises(ValueError, match="Invalid transformer connection string: 'yyn'"):
        parse_trafo_connection("yyn")
    with pytest.raises(ValueError, match="Invalid transformer connection string: 'YZN'"):
        parse_trafo_connection("YZN")
    with pytest.raises(ValueError, match="Invalid transformer connection string: 'YNx'"):
        parse_trafo_connection("YNx")
    with pytest.raises(ValueError, match="Invalid transformer connection string: 'Dy13'"):
        parse_trafo_connection("Dy13")


def test_parse_trafo3_connection__pos():
    assert parse_trafo3_connection("Dyn1yn1") == {
        "winding_1": "D",
        "winding_2": "yn",
        "clock_12": "1",
        "winding_3": "yn",
        "clock_13": "1",
    }
    assert parse_trafo3_connection("Yyn2d4") == {
        "winding_1": "Y",
        "winding_2": "yn",
        "clock_12": "2",
        "winding_3": "d",
        "clock_13": "4",
    }
    assert parse_trafo3_connection("Yzn4y0") == {
        "winding_1": "Y",
        "winding_2": "zn",
        "clock_12": "4",
        "winding_3": "y",
        "clock_13": "0",
    }
    assert parse_trafo3_connection("YNd1z0") == {
        "winding_1": "YN",
        "winding_2": "d",
        "clock_12": "1",
        "winding_3": "z",
        "clock_13": "0",
    }
    assert parse_trafo3_connection("Dy1y5") == {
        "winding_1": "D",
        "winding_2": "y",
        "clock_12": "1",
        "winding_3": "y",
        "clock_13": "5",
    }
    assert parse_trafo3_connection("Dy5d11") == {
        "winding_1": "D",
        "winding_2": "y",
        "clock_12": "5",
        "winding_3": "d",
        "clock_13": "11",
    }
    assert parse_trafo3_connection("Dyn-1yn-1") == {
        "winding_1": "D",
        "winding_2": "yn",
        "clock_12": "-1",
        "winding_3": "yn",
        "clock_13": "-1",
    }


def test_parse_trafo3_connection__neg():
    with pytest.raises(ValueError, match="Invalid three winding transformer connection string: 'Xynd'"):
        parse_trafo3_connection("Xynd")
    with pytest.raises(ValueError, match="Invalid three winding transformer connection string: 'ydyn'"):
        parse_trafo3_connection("ydyn")
    with pytest.raises(ValueError, match="Invalid three winding transformer connection string: 'DYZN'"):
        parse_trafo3_connection("DYZN")
    with pytest.raises(ValueError, match="Invalid three winding transformer connection string: 'YNxd'"):
        parse_trafo3_connection("YNxd")
    with pytest.raises(ValueError, match="Invalid three winding transformer connection string: 'Dyd13'"):
        parse_trafo3_connection("Dyd13")
    with pytest.raises(ValueError, match="Invalid three winding transformer connection string: 'DyD10'"):
        parse_trafo3_connection("DyD10")
    with pytest.raises(ValueError, match=re.escape("Invalid three winding transformer connection string: 'Dyn+5d-1'")):
        parse_trafo3_connection("Dyn+5d-1")
    with pytest.raises(ValueError, match="Invalid three winding transformer connection string: 'Dy1d13'"):
        parse_trafo3_connection("Dy1d13")
    with pytest.raises(ValueError, match="Invalid three winding transformer connection string: 'Da1d10'"):
        parse_trafo3_connection("Da1d10")
    with pytest.raises(ValueError, match="Invalid three winding transformer connection string: 'Xy1d10'"):
        parse_trafo3_connection("Xy1d10")


def test_parse_node_ref__pos():
    assert parse_node_ref("node")
    assert parse_node_ref("from_node")
    assert parse_node_ref("to_node")
    assert parse_node_ref("node_1")
    assert parse_node_ref("node_2")
    assert parse_node_ref("node_3")


def test_parse_node_ref__neg():
    with pytest.raises(ValueError, match="Invalid node reference string: 'from_bla'"):
        parse_node_ref("from_bla")
    with pytest.raises(ValueError, match="Invalid node reference string: 'a_node_node_b'"):
        parse_node_ref("a_node_node_b")
    with pytest.raises(ValueError, match="Invalid node reference string: 'nodes'"):
        parse_node_ref("nodes")
    with pytest.raises(ValueError, match="Invalid node reference string: 'anode'"):
        parse_node_ref("anode")
    with pytest.raises(ValueError, match="Invalid node reference string: 'immunodeficient'"):
        parse_node_ref("immunodeficient")


def test_is_node_ref__pos():
    assert is_node_ref("node")
    assert is_node_ref("from_node")
    assert is_node_ref("to_node")
    assert is_node_ref("node_1")
    assert is_node_ref("node_2")
    assert is_node_ref("node_3")


def test_is_node_ref__neg():
    assert not is_node_ref("from_bla")
    assert not is_node_ref("a_node_node_b")
    assert not is_node_ref("nodes")
    assert not is_node_ref("anode")
    assert not is_node_ref("immunodeficient")



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

from pydantic import TypeAdapter

from power_grid_model_io.data_types import ExtraInfo, StructuredData


def test_extra_info():
    extra_info = {1: {"a": 123, "b": 1.23}, 2: {"c": (1.2, 3.4, 5.6), "d": "foo"}}

    # Expect no exception
    adapter = TypeAdapter(ExtraInfo)
    adapter.validate_python(extra_info)


def test_structured_data__single():
    data = {"node": [{"id": 1}, {"id": 2}], "line": [{"id": 3, "node_from": 1, "node_to": 2}]}

    # Expect no exception
    adapter = TypeAdapter(StructuredData)
    assert isinstance(adapter.validate_python(data), dict)


def test_structured_data__batch():
    data = [
        {
            "load": [
                {"id": 1, "p": 111.0},
                {"id": 2, "p": 222.0},
            ]
        },
        {
            "load": [
                {"id": 1, "p": 333.0},
                {"id": 2, "p": 444.0},
            ]
        },
    ]

    # Expect no exception
    adapter = TypeAdapter(StructuredData)
    assert isinstance(adapter.validate_python(data), list)



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0
"""
Field multiplier helper class
"""

from power_grid_model_io.mappings.field_mapping import FieldMapping

Multipliers = dict[str, float]


# pylint: disable=too-few-public-methods
class MultiplierMapping(FieldMapping[float]):
    """
    Field multiplier helper class
    """

    def __init__(self, mapping: Multipliers | None = None, logger=None):
        super().__init__(mapping=mapping, logger=logger)
        if mapping is not None:
            self._log.debug("Set multiplier mapping", n_attributes=len(mapping))

    def get_multiplier(self, attr: str, table: str | None = None) -> float:
        """
        Find the multiplier for a given attribute.
        """
        return self._get_mapping(attr=attr, table=table)



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0
import shutil
import tempfile
from pathlib import Path
from unittest.mock import MagicMock, patch

import pytest
import structlog.testing

from power_grid_model_io.utils.zip import _get_only_item_in_dir, extract
from tests.utils import MockTqdm, assert_log_exists

DATA_DIR = Path(__file__).parents[2] / "data" / "zip"
ZIP1 = DATA_DIR / "foo.zip"
ZIP2 = DATA_DIR / "foo-bar.zip"


@pytest.fixture
def temp_dir():
    with tempfile.TemporaryDirectory() as tmp:
        yield Path(tmp).resolve()


@patch("power_grid_model_io.utils.download.tqdm")
def test_extract(mock_tqdm: MagicMock, temp_dir: Path):
    # Arrange
    src_file_path = temp_dir / "compressed.zip"
    dst_dir_path = temp_dir / "extracted"
    shutil.copyfile(ZIP2, src_file_path)
    mock_tqdm.side_effect = MockTqdm

    # Act
    extract_dir_path = extract(src_file_path=src_file_path, dst_dir_path=dst_dir_path)

    # Assert
    assert extract_dir_path == dst_dir_path
    assert (dst_dir_path / "foo.txt").is_file()
    assert (dst_dir_path / "folder/bar.txt").is_file()


@patch("power_grid_model_io.utils.download.tqdm")
def test_extract__auto_dir(mock_tqdm: MagicMock, temp_dir: Path):
    # Arrange
    src_file_path = temp_dir / "compressed.zip"
    shutil.copyfile(ZIP2, src_file_path)
    mock_tqdm.side_effect = MockTqdm

    # Act
    extract_dir_path = extract(src_file_path=src_file_path)

    # Assert
    assert extract_dir_path == temp_dir / "compressed"
    assert (temp_dir / "compressed" / "foo.txt").is_file()
    assert (temp_dir / "compressed" / "folder" / "bar.txt").is_file()


def test_extract__invalid_file_extension():
    # Act / Assert
    with tempfile.TemporaryDirectory() as tmp:
        invalid_file = Path(tmp) / "tempfile.download"
        invalid_file.write_text("")

        with pytest.raises(ValueError, match=r"Only files with \.zip extension are supported, got tempfile\.download"):
            extract(src_file_path=invalid_file)


def test_extract__invalid_dst_dir(temp_dir: Path):
    # Arrange
    with (temp_dir / "notadir.txt").open("wb"):
        pass

    # Act / Assert
    with pytest.raises(NotADirectoryError, match=r"Destination dir .*notadir\.txt exists and is not a directory"):
        extract(src_file_path=Path("file.zip"), dst_dir_path=temp_dir / "notadir.txt")


@patch("power_grid_model_io.utils.download.tqdm")
def test_extract__file_exists(mock_tqdm: MagicMock, temp_dir: Path):
    # Arrange
    src_file_path = temp_dir / "compressed.zip"
    dst_dir_path = temp_dir / "extracted"
    shutil.copyfile(ZIP2, src_file_path)
    mock_tqdm.side_effect = MockTqdm

    dst_dir_path.mkdir()
    with (dst_dir_path / "foo.txt").open("wb") as fp:
        fp.write(b"\0")

    # Act / Assert
    with pytest.raises(FileExistsError, match=r"Destination file .*foo\.txt exists and is not empty"):
        extract(src_file_path=src_file_path, dst_dir_path=dst_dir_path)


@patch("power_grid_model_io.utils.download.tqdm")
def test_extract__skip_if_exists(mock_tqdm: MagicMock, temp_dir: Path):
    # Arrange
    src_file_path = temp_dir / "compressed.zip"
    dst_dir_path = temp_dir / "compressed"
    shutil.copyfile(ZIP2, src_file_path)
    mock_tqdm.side_effect = MockTqdm

    dst_dir_path.mkdir()
    with (dst_dir_path / "foo.txt").open("wb") as fp:
        fp.write(b"\0")

    # Act / Assert
    with structlog.testing.capture_logs() as capture:
        extract(src_file_path=src_file_path, dst_dir_path=dst_dir_path, skip_if_exists=True)
        assert_log_exists(
            capture, "debug", "Skip file extraction, destination file exists", dst_file_path=dst_dir_path / "foo.txt"
        )


@patch("power_grid_model_io.utils.download.tqdm")
def test_extract__return_subdir_path(mock_tqdm: MagicMock, temp_dir: Path):
    # Arrange
    src_file_path = temp_dir / "foo.zip"
    shutil.copyfile(ZIP1, src_file_path)
    mock_tqdm.side_effect = MockTqdm

    # Act
    extract_dir_path = extract(src_file_path=src_file_path)

    # Assert
    assert extract_dir_path == temp_dir / "foo" / "foo"
    assert (temp_dir / "foo" / "foo" / "foo.txt").is_file()


def test_get_only_item_in_dir__no_items(temp_dir):
    # Act / Assert
    assert _get_only_item_in_dir(temp_dir) is None


def test_get_only_item_in_dir__one_file(temp_dir):
    # Arrange
    with (temp_dir / "file.txt").open("wb"):
        pass

    # Act / Assert
    assert _get_only_item_in_dir(temp_dir) == temp_dir / "file.txt"


def test_get_only_item_in_dir__one_dir(temp_dir):
    # Arrange
    (temp_dir / "subdir").mkdir()

    # Act / Assert
    assert _get_only_item_in_dir(temp_dir) == temp_dir / "subdir"


def test_get_only_item_in_dir__two_files(temp_dir):
    # Arrange
    with (temp_dir / "file_1.txt").open("wb"):
        pass
    with (temp_dir / "file_2.txt").open("wb"):
        pass

    # Act / Assert
    assert _get_only_item_in_dir(temp_dir) is None



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

from unittest.mock import MagicMock, patch

import numpy as np
import pandas as pd
import pytest
from structlog.testing import capture_logs

from power_grid_model_io.data_types import TabularData
from power_grid_model_io.mappings.unit_mapping import UnitMapping
from power_grid_model_io.mappings.value_mapping import ValueMapping
from tests.utils import assert_log_exists


@pytest.fixture
def nodes() -> pd.DataFrame:
    return pd.DataFrame([(0, 150e3), (1, 10.5e3), (2, 400.0)], columns=["id", "u_rated"])


@pytest.fixture
def nodes_iso() -> pd.DataFrame:
    return pd.DataFrame(
        [(0, 150e3), (1, 10.5e3), (2, 400.0)], columns=pd.MultiIndex.from_tuples((("id", ""), ("u_rated", "V")))
    )


@pytest.fixture
def nodes_kv() -> pd.DataFrame:
    return pd.DataFrame(
        [(0, 150), (1, 10.5), (2, 0.4)], columns=pd.MultiIndex.from_tuples((("id", ""), ("u_rated", "kV")))
    )


@pytest.fixture
def nodes_vl() -> pd.DataFrame:
    return pd.DataFrame([(0, "hv"), (1, "mv"), (2, "lv")], columns=("id", "u_rated"))


@pytest.fixture
def nodes_np() -> np.ndarray:
    return np.array([(0, 150e3), (1, 10.5e3), (2, 400.0)], dtype=[("id", "i4"), ("u_rated", "f4")])


@pytest.fixture
def lines() -> pd.DataFrame:
    return pd.DataFrame([(2, 0, 1), (3, 2, 3)], columns=("id", "from_node", "to_node"))


def test_constructor__invalid_type():
    # Act / Assert
    with pytest.raises(TypeError, match=r"Invalid.*foo.*list"):
        TabularData(foo=[])  # type: ignore


def test_get_column(nodes: pd.DataFrame, lines: pd.DataFrame):
    # Arrange
    data = TabularData(nodes=nodes, lines=lines)

    # Act
    col_data = data.get_column(table_name="nodes", column_name="u_rated")

    # Assert
    pd.testing.assert_series_equal(col_data, pd.Series([150e3, 10.5e3, 400.0], name="u_rated"))


def test_get_column__numpy(nodes_np: np.ndarray):
    # Arrange
    data = TabularData(nodes=nodes_np)

    # Act
    col_data = data.get_column(table_name="nodes", column_name="u_rated")

    # Assert
    isinstance(col_data, pd.Series)
    pd.testing.assert_series_equal(col_data, pd.Series([150e3, 10.5e3, 400.0], name="u_rated", dtype=np.float32))


def test_get_column__numpy_is_a_reference(nodes_np: np.ndarray):
    # Arrange
    data = TabularData(nodes=nodes_np)

    # Act
    col_data = data.get_column(table_name="nodes", column_name="u_rated")
    col_data.loc[0] = 123.0  # << should update the source data

    # Assert
    np.testing.assert_array_equal(nodes_np["u_rated"], np.array([123.0, 10.5e3, 400.0]))


def test_get_column__iso(nodes_iso: pd.DataFrame, lines: pd.DataFrame):
    # Arrange
    data = TabularData(nodes=nodes_iso, lines=lines)
    data.set_unit_multipliers(UnitMapping({"V": None}))

    # Act
    col_data = data.get_column(table_name="nodes", column_name="u_rated")

    # Assert
    pd.testing.assert_series_equal(col_data, pd.Series([150e3, 10.5e3, 400.0], name=("u_rated", "V")))


def test_get_column__iso_exception(nodes_iso: pd.DataFrame, lines: pd.DataFrame):
    # Arrange
    data = TabularData(nodes=nodes_iso, lines=lines)

    # Act / Assert
    with pytest.raises(KeyError, match=r"V.+u_rated.+nodes"):
        data.get_column(table_name="nodes", column_name="u_rated")


def test_get_column__unit(nodes_kv: pd.DataFrame, lines: pd.DataFrame):
    # Arrange
    data = TabularData(nodes=nodes_kv, lines=lines)
    data.set_unit_multipliers(UnitMapping({"V": {"kV": 1e3}}))

    # Act
    col_data = data.get_column(table_name="nodes", column_name="u_rated")

    # Assert
    pd.testing.assert_series_equal(col_data, pd.Series([150e3, 10.5e3, 400.0], name=("u_rated", "V")))


def test_get_column__missing_unit(nodes_kv: pd.DataFrame, lines: pd.DataFrame):
    # Arrange
    data = TabularData(nodes=nodes_kv, lines=lines)

    # Act / Assert
    with pytest.raises(KeyError, match=r"kV.+u_rated.+nodes"):
        data.get_column(table_name="nodes", column_name="u_rated")


def test_get_column__invalid_multiplier(nodes_kv: pd.DataFrame, lines: pd.DataFrame):
    # Arrange
    data = TabularData(nodes=nodes_kv, lines=lines)
    data.set_unit_multipliers(UnitMapping({"V": {"kV": "1000x"}}))  # type: ignore

    # Act / Assert
    with pytest.raises(TypeError, match=r"1000x.*kV.*not numerical"):
        data.get_column(table_name="nodes", column_name="u_rated")


def test_get_column__non_numerical_value(lines: pd.DataFrame):
    # Arrange
    nodes_txt = pd.DataFrame(
        [(0, "150 kV"), (1, "10.5 kV"), (2, "0.4 kV")],
        columns=pd.MultiIndex.from_tuples([("id", ""), ("u_rated", "kV")]),
    )
    data = TabularData(nodes=nodes_txt, lines=lines)
    data.set_unit_multipliers(UnitMapping({"V": {"kV": 1e3}}))

    # Act / Assert
    with capture_logs() as cap_log:
        col_data = data.get_column(table_name="nodes", column_name="u_rated")
    pd.testing.assert_series_equal(col_data, pd.Series(["150 kV", "10.5 kV", "0.4 kV"], name=("u_rated", "")))
    assert_log_exists(cap_log, "error", "Failed to apply unit conversion; the column is not numerical.")


def test_get_column__no_unit_conversion(nodes_kv: pd.DataFrame, lines: pd.DataFrame):
    # Arrange
    data = TabularData(nodes=nodes_kv, lines=lines)

    # Act
    col_data = data.get_column(table_name="nodes", column_name="id")

    # Assert
    pd.testing.assert_series_equal(col_data, pd.Series([0, 1, 2], name="id"))


def test_get_column__substitution(nodes_vl: pd.DataFrame, lines: pd.DataFrame):
    # Arrange
    data = TabularData(nodes=nodes_vl, lines=lines)
    data.set_substitutions(ValueMapping({"u_rated": {"mv": 10500.0, "lv": 400.0, "hv": 150000.0}}))

    # Act
    col_data = data.get_column(table_name="nodes", column_name="u_rated")

    # Assert
    pd.testing.assert_series_equal(col_data, pd.Series([150e3, 10.5e3, 400.0], name="u_rated"))


def test_get_column__substitution__numpy(nodes_np: np.ndarray):
    # Arrange
    data = TabularData(nodes=nodes_np)
    data.set_substitutions(ValueMapping({"id": {0: 100, 1: 101, 2: 102}}))

    # Act
    col_data = data.get_column(table_name="nodes", column_name="id")

    # Assert
    pd.testing.assert_series_equal(col_data, pd.Series([100, 101, 102], name="id"))


def test_get_column__substitution_exception(nodes_vl: pd.DataFrame, lines: pd.DataFrame):
    # Arrange
    data = TabularData(nodes=nodes_vl, lines=lines)
    data.set_substitutions(ValueMapping({"u_rated": {"MV": 10500.0, "LV": 400.0, "HV": 150000.0}}))

    # Act / Assert
    with pytest.raises(KeyError, match=r"hv.+u_rated.+nodes"):
        data.get_column(table_name="nodes", column_name="u_rated")


def test_get_column__index(nodes: pd.DataFrame, lines: pd.DataFrame):
    # Arrange
    data = TabularData(nodes=nodes, lines=lines)

    # Act
    nodes_idx = data.get_column(table_name="nodes", column_name="index")
    lines_idx = data.get_column(table_name="lines", column_name="index")

    # Assert
    pd.testing.assert_series_equal(nodes_idx, pd.Series([0, 1, 2], name="index"))
    pd.testing.assert_series_equal(lines_idx, pd.Series([0, 1], name="index"))


@patch("power_grid_model_io.data_types.tabular_data.TabularData._apply_unit_conversion")
def test_get_column__sanity_check(mock_unit_conversion: MagicMock, nodes_iso: pd.DataFrame, nodes_kv: pd.DataFrame):
    # Arrange
    data = TabularData(nodes=nodes_kv)
    mock_unit_conversion.return_value = nodes_iso["u_rated"]

    # Act / Assert
    with pytest.raises(TypeError, match=r"u_rated[\s\S]*unitless[\s\S]*V"):
        data.get_column(table_name="nodes", column_name="u_rated")


def test_lazy_loading(nodes: pd.DataFrame, lines: pd.DataFrame):
    # Arrange
    def nodes_fn():
        return nodes

    def lines_fn():
        return lines

    data = TabularData(nodes=nodes_fn, lines=lines_fn)

    # Assert: No data loaded
    assert not isinstance(data._data["nodes"], pd.DataFrame)
    assert not isinstance(data._data["lines"], pd.DataFrame)

    # Act / Assert: Node data loaded
    pd.testing.assert_frame_equal(data["nodes"], nodes)
    assert isinstance(data._data["nodes"], pd.DataFrame)
    assert not isinstance(data._data["lines"], pd.DataFrame)

    # Act / Assert: Line data loaded
    pd.testing.assert_frame_equal(data["lines"], lines)
    assert isinstance(data._data["nodes"], pd.DataFrame)
    assert isinstance(data._data["lines"], pd.DataFrame)


def test_len(nodes: pd.DataFrame, lines: pd.DataFrame):
    # Arrange
    data = TabularData(nodes=nodes, lines=lines)

    # Act
    n_tables = len(data)

    # Assert
    assert n_tables == 2


def test_len__empty():
    # Arrange
    data = TabularData()

    # Act
    n_tables = len(data)

    # Assert
    assert n_tables == 0


def test_contains(nodes: pd.DataFrame, lines: pd.DataFrame):
    # Arrange
    data = TabularData(nodes=nodes, lines=lines)

    # Act / Assert
    assert "nodes" in data
    assert "lines" in data


def test_contains__empty():
    # Arrange
    data = TabularData()

    # Act / Assert
    assert "nodes" not in data


def test_keys(nodes: pd.DataFrame, lines: pd.DataFrame):
    # Arrange
    data = TabularData(nodes=nodes, lines=lines)

    # Act
    keys = list(data.keys())

    # Assert
    assert keys == ["nodes", "lines"]


def test_keys__empty():
    # Arrange
    data = TabularData()

    # Act
    keys = list(data.keys())

    # Assert
    assert keys == []


def test_items(nodes: pd.DataFrame, lines: pd.DataFrame):
    # Arrange
    data = TabularData(nodes=nodes, lines=lines)

    # Act
    items = list(data.items())

    # Assert
    assert items == [("nodes", nodes), ("lines", lines)]


def test_items__empty():
    # Arrange
    data = TabularData()

    # Act
    items = list(data.items())

    # Assert
    assert items == []



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

"""
These functions can be used in the mapping files to apply functions to tabular data
"""

from power_grid_model_io.functions._functions import (
    both_zeros_to_nan,
    complex_inverse_imaginary_part,
    complex_inverse_real_part,
    degrees_to_clock,
    get_winding,
    has_value,
    is_greater_than,
    value_or_default,
    value_or_zero,
)

__all__ = [
    "both_zeros_to_nan",
    "complex_inverse_imaginary_part",
    "complex_inverse_real_part",
    "degrees_to_clock",
    "get_winding",
    "has_value",
    "is_greater_than",
    "value_or_default",
    "value_or_zero",
]
